/*
  ratbot-slam (c) 2006 Kris Beevers

  This file is part of ratbot-slam.

  ratbot-slam is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  ratbot-slam is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ratbot-slam; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
  USA
*/

// fixed-point implementation of split-merge line segment extraction
// from multiscan data

// this needs to be pretty fast; ideally we'd like to be able to do it
// multiple times per multiscan (per-strata)

#include "ratbot-slam.h"
#include <string.h>

typedef struct
{
  int start, end;
} range;

static extent_feature working_segments[MAX_FEATURES_PER_SCAN];
static range          working_ranges[MAX_FEATURES_PER_SCAN];
static extent_feature done_segments[MAX_FEATURES_PER_SCAN];
static range          done_ranges[MAX_FEATURES_PER_SCAN];

int my_sign(fixed_t f) { return (f < 0) ? -1 : 1; }

// fits a line to the points in the range ranges[range] and stores the
// resulting line in working_segments[range].  simple unweighted least
// squares estimation (no covariance estimation); borrowed mainly from
// GSL, aside from the transformation at the end.
void least_squares_fit(const fixed_t *x, const fixed_t *y, const range *ranges, int range)
{
  fixed_t m_x = 0, m_y = 0, m_dx2 = 0, m_dy2 = 0, m_dxdy = 0;
  int i, j;

  const fixed_t n = int2fp(ranges[range].end - ranges[range].start);

  for(i = ranges[range].start, j = 0; i < ranges[range].end; ++i, ++j) {
    m_x += x[i];  // FIXME: this may overflow!  (recursive approach
		  // like in slam3 resolves overflow but is
		  // computationally not as good)
    m_y += y[i];
  }
  m_x = fp_div(m_x, n);
  m_y = fp_div(m_y, n);

  for(i = ranges[range].start, j = 0; i < ranges[range].end; ++i, ++j) {
    const fixed_t dx = x[i] - m_x;
    const fixed_t dy = y[i] - m_y;
    m_dx2 += fp_mul(dx,dx);  // FIXME: this may overflow!
    m_dy2 += fp_mul(dy,dy);
    m_dxdy += fp_mul(dx,dy);
  }
  //  m_dx2 = fp_div(m_dx2, n);
  //  m_dxdy = fp_div(m_dxdy, n);

  // y=mx+b form
  fixed_t m; // avoid precision issues by using the larger of m_dx^2, m_dy^2
  if(m_dx2 > fp_1_2)
    m = fp_div(m_dxdy, m_dx2);
  else
    m = (m_dxdy == 0) ? fp_max_value : fp_div(m_dy2, m_dxdy);
  const fixed_t b = m_y - fp_mul(m_x, m);

  // transform to r=x*cos(theta) + y*sin(theta) form
  if(m > 11796480 || m < -11796480) // prevent overflow in denominator (the +1 won't make much diff anyway)
    working_segments[range].r = -fp_div(my_sign(m)*b, m);
  else
    working_segments[range].r = -fp_div(my_sign(m)*b, fp_sqrt(fp_1+fp_mul(m,m)));
  if(m != 0)
    working_segments[range].theta = -fp_atan(fp_div(fp_1,m));
  else
    working_segments[range].theta = fp_pi_2;
}

// splits all of the segments in working_segments[0..(r-1)] at the
// point associated with the segment that is farthest from the line.
// if the point is close enough, the line is not split and it is
// instead added to done_ranges.
void split(const fixed_t *x, const fixed_t *y, int *r, int *done)
{
  int i, maxi, s = 0, newr = 0;
  fixed_t d, maxd;
  static range working_ranges_buf[MAX_FEATURES_PER_SCAN];

  for(; s < *r; ++s) { // iterate over current segments

    // find farthest point from the segment
    maxd = fp_min_value;
    maxi = working_ranges[s].start;
    for(i = maxi; i < working_ranges[s].end; ++i) {
      d = squared_distance_point_line
	(x[i], y[i], working_segments[s].r, working_segments[s].theta);
      if(d > maxd) {
	maxd = d;
	maxi = i;
      }
    }

    if(maxd < LE_TMAX_2) // done with this segment
      done_ranges[(*done)++] = working_ranges[s];
    else { // split into two segments
      if(maxi == working_ranges[s].start)
	++maxi;
      else if(maxi == working_ranges[s].end)
	--maxi;

      if(maxi - working_ranges[s].start >= LE_MIN_POINTS) {
	working_ranges_buf[newr].start = working_ranges[s].start;
	working_ranges_buf[newr].end = maxi;
	++newr;
      }
      if(working_ranges[s].end - maxi >= LE_MIN_POINTS) {
	working_ranges_buf[newr].start = maxi;
	working_ranges_buf[newr].end = working_ranges[s].end;
	++newr;
      }
    }
  }

  // FIXME: this is not very nice; it might be better to have a
  // pointer that switches between two separate buffers indicating the
  // one currently in use
  memcpy(working_ranges, working_ranges_buf, newr*sizeof(range));
  *r = newr;
}

// project the points associated with the line working_segments[range]
// onto the line and take the two extremal projected points as the
// segment endpoints
void set_extent(const fixed_t *x, const fixed_t *y, int range)
{
  extent_feature *s = &(working_segments[range]);
  int i = done_ranges[range].start;
  fixed_t u;

  s->umax = fp_min_value;
  s->umin = fp_max_value;
  for(; i < done_ranges[range].end; ++i) {
    u = project_point_onto_line(x[i], y[i], s->r, s->theta);
    if(u < s->umin) s->umin = u;
    if(u > s->umax) s->umax = u;
  }

  // update endpoints
  u_to_point(s->umin, s->r, s->theta, &(s->xmin), &(s->ymin));
  u_to_point(s->umax, s->r, s->theta, &(s->xmax), &(s->ymax));
}

// merge similar segments, put the results in done_segments, and set
// done to the number of output segments
void merge(int *done)
{
  fixed_t md;
  int i, j;
  extent_feature *f1, *f2;
  static int     matches[MAX_FEATURES_PER_SCAN];
  static fixed_t dists[MAX_FEATURES_PER_SCAN];



  //FIXME REMOVE
  for(i=0;i<*done;++i)
    done_segments[i] = working_segments[i];
  return;



  // initialization
  for(i = 0; i < *done; ++i) {
    dists[i] = fp_max_value;
    matches[i] = i;
  }

  // compute pairwise squared mahalanobis distances for overlapping
  // segments and get the nearest neighbors
  for(i = 0; i < *done; ++i) {
    f1 = &(working_segments[i]);
    for(j = i+1; j < *done; ++j) {
      f2 = &(working_segments[j]);
      if(segments_overlap(f1, f2)) {
	md = squared_mahalanobis(f1, f2);
	if(md < dists[i]) { dists[i] = md; matches[i] = j; }
	if(md < dists[j]) { dists[j] = md; matches[j] = i; }
      }
    }
  }

  // if ML pairings are edges in a graph, then we'll merge all
  // features in the same connected component of the graph into a
  // single feature
  for(i = 0, j = 0; i < *done; ++i) {
    if(matches[i] < i && dists[i] < LE_MERGE_THRESH_2) {
      matches[i] = matches[matches[i]];
      merge_segments(&(done_segments[matches[i]]), &(working_segments[i]));
    } else {
      done_segments[j] = working_segments[i];
      matches[i] = j++;
    }
  }
}


/// FIXME FIXME: a problem with this is that it doesn't directly make
/// use of per-point uncertainty which means it's hard for us to
/// inflate the uncertainty of the extracted segments based on
/// uncertainty of the poses they were observed from.  we could store
/// separate covariances for every point (adding in pose uncertainty),
/// and then compute from these the range-bearing covariances.  this
/// is kind of ugly but it may be necessary to deal with uncertainty
/// in the expected multiscan trajectory.

// estimates line parameters and covariance for points in the range
// ranges[range], and stores the resulting line in
// working_segments[range].  assumes that an inital least squares fit
// has been performed to get an initial estimate of the line
// parameters
void ml_line_estimate
  (const fixed_t *x, const fixed_t *y, const fixed_t *u,
   const range *ranges, int range)
{
  int k;
  static fixed_t ck, sk, dk_2, dk, phik, dk_ck, dk_sk, ck_2, sk_2, dk_2_sk_2,
    two_dk_2_bb_minus_rr, one_over_bk, Prr, ak, akp, akpp, bk, bkp, bkpp, GTpp, den;

  // CHEAT: in general, the difference between the maximum likelihood
  // estimate and the least squares estimate is quite small.  on a
  // fixed-point platform, the error in the number representation is
  // almost surely enough to overshadow the difference.  so instead of
  // using a time-consuming iterative approach for accurate (r,theta)
  // line parameter estimation, we'll just use the least squares
  // estimate for (r,theta) and compute the covariance based on this.

  least_squares_fit(x, y, ranges, range);

  // compute covariance (mostly) according to the approach of:
  // "Weighted line fitting algorithms for mobile robot map building
  // and efficient data representation."  Pfister, Roumeliotis,
  // Burdick, ICRA03.  this is in fact an approximation because it
  // makes the assumption that all of the readings came from the same
  // pose, whereas in our case they do not.  however, they all came
  // from nearby poses and the approximation is pretty good in
  // practice.

  // FIXME: how to inflate covariance with the "scalar uncertainties"
  // for each point?

  extent_feature *f = &(working_segments[range]);


  /*
  // FIXME REMOVE HERE HERE FAKE COVARIANCE
  float n = (float)(ranges[range].end-ranges[range].start);
  f->rr = float2fp(1.0/n);
  f->rt = 0;
  f->tt = float2fp(1.0/n);
  return;
  */


  GTpp = 0;
  Prr = 0;
  f->rr = f->rt = f->tt = 0;
  for(k = ranges[range].start; k < ranges[range].end; ++k) { // iterate over points

    // first compute range/bearing wrt pose (0,0,0) --- note that this
    // means the points need to be given to us transformed so that the
    // reference frame of the robot is at the origin
    dk_2 = fp_mul(x[k],x[k]) + fp_mul(y[k],y[k]);
    dk = fp_sqrt(dk_2);
    phik = fp_atan2(y[k], x[k]);
    ck = fp_cos(f->theta - phik);
    sk = fp_sin(f->theta - phik);
    ck_2 = fp_mul(ck, ck);
    sk_2 = fp_mul(sk, sk);
    dk_ck = fp_mul(dk, ck);
    dk_sk = fp_mul(dk, sk);
    dk_2_sk_2 = fp_mul(dk_sk, dk_sk);
    two_dk_2_bb_minus_rr = fp_mul(fp_2, fp_mul(dk_2, VAR_BEARING) - VAR_RANGE);

    // now compute all the necessary coefficients to obtain the
    // covariance
    ak = dk_ck - f->r;
    akp = fp_mul(fp_mul(-fp_2, dk_sk), ak);
    akpp = fp_mul(-fp_2, dk_2_sk_2) - fp_mul(fp_mul(fp_2, dk_ck), ak);
    ak = fp_mul(ak, ak);

    bk = fp_mul(VAR_RANGE, ck_2) + fp_mul(VAR_BEARING, dk_2_sk_2);
    bkp = fp_mul(fp_mul(ck, sk), two_dk_2_bb_minus_rr);
    bkpp = fp_mul(two_dk_2_bb_minus_rr, ck_2 - sk_2);
    one_over_bk = fp_div(fp_1, bk);
    den = fp_mul(bk, fp_mul(bk, bk));

    GTpp += fp_div( (fp_mul(fp_mul(akpp,bk)-fp_mul(ak,bkpp), bk) -
		     fp_mul(fp_mul(fp_2, bkp), fp_mul(akp,bk)-fp_mul(ak,bkp))),
		    (den == 0) ? 1 : den);

    Prr += one_over_bk;
    f->rt += fp_mul(fp_mul(fp_2, dk_sk), one_over_bk);
    f->tt += fp_mul(fp_mul(fp_2+fp_2, dk_2_sk_2), one_over_bk);
  }

  // avert underflow (HACK)
  if(Prr == 0) Prr = 1;
  if(GTpp == 0) GTpp = 1;
  den = fp_mul(GTpp, GTpp);
  if(den == 0) den = 1;

  // compute the final covariance
  f->rr = Prr > 65535 ? 1 : (Prr < 0 ? 1 : fp_div(fp_1, Prr));
  f->rt = fp_mul(f->rt, GTpp > 65535 ? 1 : fp_div(f->rr, GTpp));
  f->tt = den > 65535 ? 1 : (den < 0 ? 1 : fp_div(f->tt, den));

#if 0
  if(f->tt < 0 || f->rr < 0)
    printf("WTF\n");
#endif
}

void extract_line_segments
  (const fixed_t *x, const fixed_t *y, const fixed_t *u, int n,
   extent_feature *out, int *num_features)
{
  int r = 0, i = 0, done = 0;

#if 0
  if(n) printf("\n");
  for(; i < n; ++i) {
    printf("%g %g\n", fp2float(x[i]), fp2float(y[i]));
  } i = 0;
  //  return;
#endif

  // initial clustering (breakpoint detection)
  working_ranges[r].start = 0;
  for(; i < n; ++i) {
    if(squared_distance_point_point(x[i], y[i], x[i-1], y[i-1]) > LE_CLUSTER_THRESH_2) {
      if(i - working_ranges[r].start > LE_MIN_POINTS) {
	working_ranges[r].end = i; // keep this point set
	++r;
      }
      working_ranges[r].start = i; // start a new point set
    }
  }
  if(i - working_ranges[r].start > LE_MIN_POINTS) {
    working_ranges[r].end = i;
    ++r;
  }

  // fit, split, repeat until convergence
  while(r > 0) {
    for(i = 0; i < r; ++i) {
      least_squares_fit(x, y, working_ranges, i);
      //                  printf("%d -- %d: %g %g\n", working_ranges[i].start, working_ranges[i].end,
      //            	     fp2float(working_segments[i].r), fp2float(working_segments[i].theta));
    }
    split(x, y, &r, &done); // done gets # of working_segments/done_ranges
  }
  //  printf("\n");

  // obtain full maximum likelihood parameter and covariance
  // estimates, and also segment endpoints
  for(i = 0; i < done; ++i) {
    ml_line_estimate(x, y, u, done_ranges, i);
    set_extent(x, y, i);
  }

  // merge similar segments
  merge(&done);

  // discard too-short segments and return the rest
  for(*num_features = 0, i = 0; i < done; ++i)
    if(squared_length(&done_segments[i]) >= LE_MIN_LENGTH_2) {
      out[(*num_features)++] = done_segments[i];

#if 0
      if(done_segments[i].rr < 0 || done_segments[i].tt < 0) {
	printf("!!?!\n");
      }
#endif

#if 0
      //      printf("%g %g\n", fp2float(done_segments[i].r), fp2float(done_segments[i].theta));
      #if 0
      int k=done_ranges[i].start;
      for(;k<done_ranges[i].end;++k) {
	printf("%g %g\n", fp2float(x[k]), fp2float(y[k]));
      }
      #endif
      printf("%g %g %g %g\n\n", fp2float(done_segments[i].xmin), fp2float(done_segments[i].ymin),
	     fp2float(done_segments[i].xmax), fp2float(done_segments[i].ymax));
      //exit(0);
#endif

    }
#if 0
    else
      printf("TOO SHORT: %g %g (%g)\n", fp2float(done_segments[i].r), fp2float(done_segments[i].theta),
	     fp2float(squared_length(&done_segments[i])));
#endif

  //  printf("\n");
}
