/*
  ratbot-slam (c) 2006 Kris Beevers

  This file is part of ratbot-slam.

  ratbot-slam is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  ratbot-slam is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with ratbot-slam; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
  USA
*/

// some math for slam with line features, mainly geometry

#include <fixedpoint.h>
#include "ratbot-slam.h"

fixed_t angular_distance(fixed_t a1, fixed_t a2)
{
  fixed_t a = fp_clamp_angle(a1-a2);
  if(a > fp_pi) a -= fp_2_pi;
  return a;
}

fixed_t squared_distance_point_point(fixed_t x1, fixed_t y1, fixed_t x2, fixed_t y2)
{
  const fixed_t dx = x2-x1, dy = y2-y1;
  return fp_mul(dx,dx) + fp_mul(dy,dy);
}

fixed_t squared_distance_point_line(fixed_t x, fixed_t y, fixed_t r, fixed_t theta)
{
  fixed_t x2, y2;
  const fixed_t u = project_point_onto_line(x, y, r, theta);
  u_to_point(u, r, theta, &x2, &y2);
  return squared_distance_point_point(x, y, x2, y2);
}

fixed_t squared_distance_point_segment(fixed_t x, fixed_t y, const extent_feature *seg)
{
  const fixed_t u = project_point_onto_line(x, y, seg->r, seg->theta);
  if(u >= seg->umin && u <= seg->umax) {
    fixed_t x2, y2;
    u_to_point(u, seg->r, seg->theta, &x2, &y2);
    return squared_distance_point_point(x, y, x2, y2);
  } else if(u < seg->umin)
    return squared_distance_point_point(x, y, seg->xmin, seg->ymin);
  else
    return squared_distance_point_point(x, y, seg->xmax, seg->ymax);
}

fixed_t squared_length(const extent_feature *seg)
{
  return squared_distance_point_point(seg->xmin, seg->ymin, seg->xmax, seg->ymax);
}

fixed_t project_point_onto_line(fixed_t x, fixed_t y, fixed_t r, fixed_t theta)
{
  return fp_mul(x - fp_mul(r, fp_cos(theta)), fp_cos(theta+fp_pi_2)) +
    fp_mul(y - fp_mul(r, fp_sin(theta)), fp_sin(theta+fp_pi_2));
}

void u_to_point(fixed_t u, fixed_t r, fixed_t theta, fixed_t *x, fixed_t *y)
{
  *x = fp_mul(r, fp_cos(theta)) + fp_mul(u, fp_cos(theta+fp_pi_2));
  *y = fp_mul(r, fp_sin(theta)) + fp_mul(u, fp_sin(theta+fp_pi_2));
}

fixed_t squared_mahalanobis(const extent_feature *f1, const extent_feature *f2)
{
  // this is a fully written out version of nu*inv(c1+c2)*nu'
  // FIXME: static-ify some variables to avoid lots of stack action?
  const fixed_t nu_r = f2->r - f1->r;
  const fixed_t nu_t = angular_distance(f2->theta, f1->theta);
  const fixed_t srr = f1->rr + f2->rr, srt = f1->rt + f2->rt, stt = f1->tt + f2->tt;
  const fixed_t den = fp_mul(srr, stt) - fp_mul(srt, srt);
  const fixed_t one_over_det = den == 0 ? fp_max_value : fp_div(fp_1, den);
  const fixed_t irr = fp_mul(stt, one_over_det),
    irt = -fp_mul(srt, one_over_det),
    itt = fp_mul(srr, one_over_det);
  if(irr < 0 || itt < 0) // HACK: simple overflow check
    return fp_max_value;
  fixed_t ret = fp_mul(irr, fp_mul(nu_r,nu_r)) + fp_mul(fp_2, fp_mul(irt, fp_mul(nu_r,nu_t))) +
    fp_mul(itt, fp_mul(nu_t,nu_t));
  if(ret < 0) // HACK: simple overflow check
    return fp_max_value;
  return ret;
}

fixed_t likelihood_from_md_2(const extent_feature *f1, const extent_feature *f2, fixed_t md_2)
{
#if 0 // the full version
  const fixed_t rr = f1->rr + f2->rr;
  const fixed_t rt = f1->rt + f2->rt;
  const fixed_t tt = f1->tt + f2->tt;
  const fixed_t d = fp_mul(fp_2_pi, fp_sqrt(fp_mul(rr,tt) - fp_mul(rt,rt)));
  return fp_div(fp_exp(-fp_mul(fp_1_2, md_2)), d);
#endif
  // an alternative that is faster and works better with fixed point
  return fp_exp(-fp_mul(fp_1_2, md_2));
}

void transform_points
  (const fixed_t *x, const fixed_t *y, int n,
   const pose_t *delta, fixed_t *xout, fixed_t *yout)
{
  const fixed_t ct = fp_cos(delta->t), st = fp_sin(delta->t);
  int i = 0;
  for(; i < n; ++i) {
    xout[i] = delta->x + fp_mul(x[i],ct) - fp_mul(y[i],st);
    yout[i] = delta->y + fp_mul(x[i],st) + fp_mul(y[i],ct);
  }
}

void transform_points_and_covs
  (const fixed_t *x, const fixed_t *y,
   const fixed_t *xx, const fixed_t *xy, const fixed_t *yy,
   int n, const pose_t *delta,
   fixed_t *xout, fixed_t *yout, fixed_t *xxout, fixed_t *xyout, fixed_t *yyout)
{
  const fixed_t ct = fp_cos(delta->t), st = fp_sin(delta->t);
  int i = 0;
  for(; i < n; ++i) {
    xout[i] = delta->x + fp_mul(x[i],ct) - fp_mul(y[i],st);
    yout[i] = delta->y + fp_mul(x[i],st) + fp_mul(y[i],ct);
    const fixed_t ct_xy = fp_mul(ct,xy[i]);
    const fixed_t st_xy = fp_mul(st,xy[i]);
    const fixed_t ct_xx_st_xy = fp_mul(ct,xx[i]) - st_xy;
    const fixed_t ct_xy_st_yy = ct_xy - fp_mul(st,yy[i]);
    xxout[i] = fp_mul(ct_xx_st_xy, ct) - fp_mul(ct_xy_st_yy, st);
    xyout[i] = fp_mul(ct_xx_st_xy, st) + fp_mul(ct_xy_st_yy, ct);
    yyout[i] = fp_mul(fp_mul(st,xx[i])+ct_xy, st) + fp_mul(st_xy+fp_mul(ct,yy[i]), ct);
  }
}

void transform_segment(const extent_feature *in, const pose_t *delta, extent_feature *out)
{
  // transform (r,theta)
  out->theta = in->theta + delta->t;
  fixed_t ct = fp_cos(out->theta), st = fp_sin(out->theta);
  out->r = in->r + fp_mul(delta->x, ct) + fp_mul(delta->y, st);

  // transform covariance
  const fixed_t dS = fp_mul(delta->y, ct) - fp_mul(delta->x, st);
  out->rr = in->rr + fp_mul(fp_2, fp_mul(dS,in->rt)) + fp_mul(dS, fp_mul(dS,in->tt));
  out->rt = in->rt + fp_mul(dS,in->tt);
  out->tt = in->tt;

  // transform endpoints and extent range
  ct = fp_cos(delta->t), st = fp_sin(delta->t);
  out->xmin = delta->x + fp_mul(in->xmin,ct) - fp_mul(in->ymin,st);
  out->ymin = delta->y + fp_mul(in->xmin,st) + fp_mul(in->ymin,ct);
  out->umin = project_point_onto_line(out->xmin, out->ymin, out->r, out->theta);
  out->xmax = delta->x + fp_mul(in->xmax,ct) - fp_mul(in->ymax,st);
  out->ymax = delta->y + fp_mul(in->xmax,st) + fp_mul(in->ymax,ct);
  out->umax = project_point_onto_line(out->xmax, out->ymax, out->r, out->theta);
}

// perhaps a more ideal approach would be to compute the merged line
// for the two features and project the segment endpoints onto that.
// instead we'll directly project the endpoints of the lines on to
// each other, which isn't vastly different but is a lot faster.
int segments_overlap(const extent_feature *f1, const extent_feature *f2)
{
  fixed_t u = project_point_onto_line(f1->xmin, f2->ymin, f2->r, f2->theta);
  if(u >= f2->umin - OVERLAP_EXTRA_LEN && u <= f2->umax + OVERLAP_EXTRA_LEN)
    return 1;
  u = project_point_onto_line(f1->xmax, f1->ymax, f2->r, f2->theta);
  if(u >= f2->umin - OVERLAP_EXTRA_LEN && u <= f2->umax + OVERLAP_EXTRA_LEN)
    return 1;
  u = project_point_onto_line(f2->xmin, f2->ymin, f1->r, f1->theta);
  if(u >= f1->umin - OVERLAP_EXTRA_LEN && u <= f1->umax + OVERLAP_EXTRA_LEN)
    return 1;
  return 0; // no overlap
}

// stores merged feature distribution in f1; this whole things is
// really just a written out kalman update:
//   K = c1*inv(c1+c2)
//   f1 = f1 + K*nu
//   c1 = c1 - K*c1
void merge_segments(extent_feature *f1, const extent_feature *f2)
{

  //return;

  // FIXME: note that we will have already computed exactly these
  // things (up to inverse cov) when computing the squared mahalanobis
  // distance!  possibly we could store them!
  const fixed_t nu_r = f2->r - f1->r;
  const fixed_t nu_t = angular_distance(f2->theta, f1->theta);
  const fixed_t srr = f1->rr + f2->rr, srt = f1->rt + f2->rt, stt = f1->tt + f2->tt;
  const fixed_t den = fp_mul(srr, stt) - fp_mul(srt, srt);
  const fixed_t one_over_det = den == 0 ? fp_max_value : fp_div(fp_1, den);
  const fixed_t irr = fp_mul(stt, one_over_det),
    irt = -fp_mul(srt, one_over_det),
    itt = fp_mul(srr, one_over_det);

  extent_feature ret = *f1;

  // compute some constants
  const fixed_t rr_rr_rt_rt = fp_mul(f1->rr, irr) + fp_mul(f1->rt, irt);
  const fixed_t rr_rt_rt_tt = fp_mul(f1->rr, irt) + fp_mul(f1->rt, itt);
  const fixed_t rt_rr_tt_rt = fp_mul(f1->rt, irr) + fp_mul(f1->tt, irt);
  const fixed_t rt_rt_tt_tt = fp_mul(f1->rt, irt) + fp_mul(f1->tt, itt);

  // update mean
  ret.r += fp_mul(nu_r, rr_rr_rt_rt) + fp_mul(nu_t, rr_rt_rt_tt);
  ret.theta += fp_mul(nu_r, rt_rr_tt_rt) + fp_mul(nu_t, rt_rt_tt_tt);
  ret.theta = fp_clamp_angle(ret.theta);

  // update covariance
  ret.rr -= fp_mul(f1->rr, rr_rr_rt_rt) + fp_mul(f1->rt, rr_rt_rt_tt);
  ret.rt -= fp_mul(f1->rt, rr_rr_rt_rt) + fp_mul(f1->tt, rr_rt_rt_tt);
  ret.tt -= fp_mul(f1->rt, rt_rr_tt_rt) + fp_mul(f1->tt, rt_rt_tt_tt);

  // update extent

  // compute extent range along line
  ret.umin = ret.umax = project_point_onto_line(f1->xmin, f1->ymin, ret.r, ret.theta);
  fixed_t u = project_point_onto_line(f1->xmax, f1->ymax, ret.r, ret.theta);
  if(u < ret.umin) ret.umin = u; else if(u > ret.umax) ret.umax = u;
  u = project_point_onto_line(f2->xmin, f2->ymin, ret.r, ret.theta);
  if(u < ret.umin) ret.umin = u; else if(u > ret.umax) ret.umax = u;
  u = project_point_onto_line(f2->xmax, f2->ymax, ret.r, ret.theta);
  if(u < ret.umin) ret.umin = u; else if(u > ret.umax) ret.umax = u;

  // update endpoints
  u_to_point(ret.umin, ret.r, ret.theta, &ret.xmin, &ret.ymin);
  u_to_point(ret.umax, ret.r, ret.theta, &ret.xmax, &ret.ymax);

#if 0
  printf("%g %g %g %g\n%g %g %g %g\n%g %g %g %g\n",
	 fp2float(f1->xmin), fp2float(f1->ymin), fp2float(f1->xmax), fp2float(f1->ymax),
	 fp2float(f2->xmin), fp2float(f2->ymin), fp2float(f2->xmax), fp2float(f2->ymax),
	 fp2float(ret.xmin), fp2float(ret.ymin), fp2float(ret.xmax), fp2float(ret.ymax));
#endif

  *f1 = ret; // copy result into f1
}

// stores merged pose distribution in p1/c1; this is also just a
// written out kalman update
void merge_poses(pose_t *p1, cov3_t *c1, const pose_t *p2, const cov3_t *c2)
{
  const fixed_t nu_x = p2->x - p1->x;
  const fixed_t nu_y = p2->y - p1->y;
  const fixed_t nu_t = angular_distance(p2->t, p1->t);
  cov3_t i = *c1;
  add_cov3(&i, c2);
  inv_cov3(&i);

  const fixed_t xx_xx_xy_xy_xt_xt = fp_mul(c1->xx,i.xx)+fp_mul(c1->xy,i.xy)+fp_mul(c1->xt,i.xt);
  const fixed_t xx_xy_xy_yy_xt_yt = fp_mul(c1->xx,i.xy)+fp_mul(c1->xy,i.yy)+fp_mul(c1->xt,i.yt);
  const fixed_t xx_xt_xy_yt_xt_tt = fp_mul(c1->xx,i.xt)+fp_mul(c1->xy,i.yt)+fp_mul(c1->xt,i.tt);
  const fixed_t xy_xx_yy_xy_yt_xt = fp_mul(c1->xy,i.xx)+fp_mul(c1->yy,i.xy)+fp_mul(c1->yt,i.xt);
  const fixed_t xy_xy_yy_yy_yt_yt = fp_mul(c1->xy,i.xy)+fp_mul(c1->yy,i.yy)+fp_mul(c1->yt,i.yt);
  const fixed_t xy_xt_yy_yt_yt_tt = fp_mul(c1->xy,i.xt)+fp_mul(c1->yy,i.yt)+fp_mul(c1->yt,i.tt);
  const fixed_t xt_xx_yt_xy_tt_xt = fp_mul(c1->xt,i.xx)+fp_mul(c1->yt,i.xy)+fp_mul(c1->tt,i.xt);
  const fixed_t xt_xy_yt_yy_tt_yt = fp_mul(c1->xt,i.xy)+fp_mul(c1->yt,i.yy)+fp_mul(c1->tt,i.yt);
  const fixed_t xt_xt_yt_yt_tt_tt = fp_mul(c1->xt,i.xt)+fp_mul(c1->yt,i.yt)+fp_mul(c1->tt,i.tt);


  // update mean
  pose_t pret;
  pret.x = p1->x + fp_mul(nu_x,xx_xx_xy_xy_xt_xt) + fp_mul(nu_y,xx_xy_xy_yy_xt_yt) +
    fp_mul(nu_t,xx_xt_xy_yt_xt_tt);
  pret.y = p1->y + fp_mul(nu_x,xy_xx_yy_xy_yt_xt) + fp_mul(nu_y,xy_xy_yy_yy_yt_yt) +
    fp_mul(nu_t,xy_xt_yy_yt_yt_tt);
  pret.t = p1->t + fp_mul(nu_x,xt_xx_yt_xy_tt_xt) + fp_mul(nu_y,xt_xy_yt_yy_tt_yt) +
    fp_mul(nu_t,xt_xt_yt_yt_tt_tt);
  pret.t = fp_clamp_angle(pret.t);

  // update covariance
  cov3_t cret;
  cret.xx = c1->xx - fp_mul(c1->xx,xx_xx_xy_xy_xt_xt) - fp_mul(c1->xy,xx_xy_xy_yy_xt_yt) -
    fp_mul(c1->xt,xx_xt_xy_yt_xt_tt);
  cret.xy = c1->xy - fp_mul(c1->xy,xx_xx_xy_xy_xt_xt) - fp_mul(c1->yy,xx_xy_xy_yy_xt_yt) -
    fp_mul(c1->yt,xx_xt_xy_yt_xt_tt);
  cret.xt = c1->xt - fp_mul(c1->xt,xx_xx_xy_xy_xt_xt) - fp_mul(c1->yt,xx_xy_xy_yy_xt_yt) -
    fp_mul(c1->tt,xx_xt_xy_yt_xt_tt);
  cret.yy = c1->yy - fp_mul(c1->xy,xy_xx_yy_xy_yt_xt) - fp_mul(c1->yy,xy_xy_yy_yy_yt_yt) -
    fp_mul(c1->yt,xy_xt_yy_yt_yt_tt);
  cret.yt = c1->yt - fp_mul(c1->xt,xy_xx_yy_xy_yt_xt) - fp_mul(c1->yt,xy_xy_yy_yy_yt_yt) -
    fp_mul(c1->tt,xy_xt_yy_yt_yt_tt);
  cret.tt = c1->tt - fp_mul(c1->xt,xt_xx_yt_xy_tt_xt) - fp_mul(c1->yt,xt_xy_yt_yy_tt_yt) -
    fp_mul(c1->tt,xt_xt_yt_yt_tt_tt);

  // copy result into p1/c1
  *p1 = pret;
  *c1 = cret;
}

// adds c2 to c1
void add_cov3(cov3_t *c1, const cov3_t *c2)
{
  c1->xx += c2->xx;
  c1->xy += c2->xy;
  c1->xt += c2->xt;
  c1->yy += c2->yy;
  c1->yt += c2->yt;
  c1->tt += c2->tt;
}

// inverts c
void inv_cov3(cov3_t *c)
{
  const fixed_t one_over_det =
    fp_div(fp_1, fp_mul(c->xx,fp_mul(c->yy,c->tt)) -
	         fp_mul(c->xx,fp_mul(c->yt,c->yt)) +
	         fp_mul(fp_2,fp_mul(c->xy,fp_mul(c->xt,c->yt))) -
	         fp_mul(c->xt,fp_mul(c->xt,c->yy)));
  cov3_t ret;
  ret.xx = fp_mul(fp_mul(c->yy,c->tt) - fp_mul(c->yt,c->yt), one_over_det);
  ret.xy = -fp_mul(fp_mul(c->xy,c->tt) - fp_mul(c->xt,c->yt), one_over_det);
  ret.xt = fp_mul(fp_mul(c->xy,c->yt) - fp_mul(c->xt,c->yy), one_over_det);
  ret.yy = fp_mul(fp_mul(c->xx,c->tt) - fp_mul(c->xt,c->xt), one_over_det);
  ret.yt = -fp_mul(fp_mul(c->xx,c->yt) - fp_mul(c->xy,c->xt), one_over_det);
  ret.tt = fp_mul(fp_mul(c->xx,c->yy) - fp_mul(c->xy,c->xy), one_over_det);
  *c = ret;
}

fixed_t gaussian(fixed_t mean, fixed_t stddev)
{
  static int cached = 0;
  static fixed_t extra;

  if(cached) {
    cached = 0;
    return fp_mul(extra, stddev) + mean;
  }
  
  // pick random point in unit circle
  static fixed_t a, b, c, t;
  do {
    a = fp_mul(fp_2, fp_rand_0_1()) - fp_1;
    b = fp_mul(fp_2, fp_rand_0_1()) - fp_1;
    c = fp_mul(a,a) + fp_mul(b,b);
  } while(c > fp_1 || c == 0);
#ifdef _USE_DOUBLE
  t = fp_sqrt(-fp_div(fp_mul(fp_2, fp_log(c)), c));
#else
  t = pgm_read_fixed(&unit_gaussian_table[c >> unit_gaussian_shift]);
#endif
  extra = fp_mul(t, a);
  cached = 1;
  return fp_mul(fp_mul(t, b), stddev) + mean;
}

#if 0
// explicitly written-out cholesky decomposition of a 3x3 symmetric
// (positive, but check for that anyway) definite matrix: find a
// triangular matrix L such that LL' = cov; this follows the
// Cholesky-Banachiewicz algorithm (see, e.g.,
// http://en.wikipedia.org/wiki/Cholesky_decomposition)
void cholesky(const cov3_t *cov, cov3_t *out)
{
  out->xx = fp_sqrt(cov->xx);
  if(out->xx > 0) {
    out->xy = fp_div(cov->xy, out->xx);
    out->xt = fp_div(cov->xt, out->xx);
  } else
    out->xy = out->xt = 0;
  // add an epsilon for numerical error (1 in fixed point is the
  // smallest possible value)
  out->yy = fp_sqrt(cov->yy - fp_mul(out->xy,out->xy) + 2);
  if(out->yy > 0)
    out->yt = fp_div(cov->yt - fp_mul(out->xt,out->xy), out->yy);
  else
    out->yt = 0;
  out->tt = fp_sqrt(cov->tt - fp_mul(out->xt,out->xt) - fp_mul(out->yt,out->yt) + 2);
}
#endif

void gaussian_pose(const pose_t *mean, const cov3_t *cov, pose_t *sample)
{
  // fast version that ignores cross-covariance terms (so the samples
  // are not as good):
  sample->x = gaussian(mean->x, fp_sqrt(cov->xx));
  sample->y = gaussian(mean->y, fp_sqrt(cov->yy));
  sample->t = gaussian(mean->t, fp_sqrt(cov->tt));

#if 0
  // sampling from a multivariate gaussian using the cholesky
  // decomposition

  // first, compute the cholesky decomposition to find an L such that
  // LL' = cov
  static cov3_t L;
  cholesky(cov, &L);

  // now draw a vector z of n (3 in this case) independent values from
  // the standard normal distribution N(0,1)
  fixed_t zx = gaussian(0, fp_1);
  fixed_t zy = gaussian(0, fp_1);
  fixed_t zt = gaussian(0, fp_1);

  sample->x = mean->x + fp_mul(L.xx,zx) + fp_mul(L.xy,zy) + fp_mul(L.xt,zt);
  sample->y = mean->y + fp_mul(L.xy,zx) + fp_mul(L.yy,zy) + fp_mul(L.yt,zt);
  sample->t = mean->t + fp_mul(L.xt,zx) + fp_mul(L.yt,zy) + fp_mul(L.tt,zt);
#endif
}

void swap(fixed_t *x, fixed_t *y, fixed_t *phi, int a, int b)
{
  if(a == b)
    return;
  static fixed_t tx, ty, tp;
  tx = x[a]; ty = y[a]; tp = phi[a];
  x[a] = x[b]; y[a] = y[b]; phi[a] = phi[b];
  x[b] = tx; y[b] = ty; phi[b] = tp;
}

const fixed_t *gphi;
int compare(const void *va, const void *vb)
{
  int a = *(int *)va;
  int b = *(int *)vb;
  return (gphi[a] <= gphi[b]) ? -1 : 1;
}

void radial_sort_points(fixed_t *x, fixed_t *y, fixed_t *phi, int n, const pose_t *origin)
{
  // first compute angles to (x,y)
  int i;
  for(i = 0; i < n; ++i)
    phi[i] = fp_atan2(y[i]-origin->y, x[i]-origin->x);
  gphi = phi; // for compare

  // now do sorting wrt phi using simple bubble sort (FIXME: something faster)
  int swapped;
  do {
    swapped = 0;
    for(i = 0; i < n-2; ++i) {
      if(phi[i] > phi[i+1]) {
	swap(x, y, phi, i, i+1);
	swapped = 1;
      }
    }
  } while(swapped);
}
